import unittest
from unittest.mock import MagicMock, patch

import os
import pathlib
from pathlib import Path

from txrm2tiff.shortcut_creator import create_Windows_shortcut, _create_lnk_file

class TestSetupFunctions(unittest.TestCase):

    @patch.object(pathlib.Path, "exists", MagicMock(return_value=False))
    @patch('logging.error')
    def test_setup_windows_shortcut_function_called(self, mocked_error):
        with patch('txrm2tiff.shortcut_creator._create_lnk_file', MagicMock()) as mocked_shortcut_creator:
            create_Windows_shortcut()
            if os.name == "nt":
                mocked_shortcut_creator.assert_called_once_with(Path(os.environ["HOMEDRIVE"]) / os.environ["HOMEPATH"] / "Desktop" / "txrm2tiff.lnk")
            else:
                mocked_error.assert_called_once_with("This command is only valid on Windows installations.")

    def test_setup_windows_shortcut_test_created(self):
        # Only run this test if on Windows
        if os.name == "nt":
            test_shortcut_path = Path(".") / "test_shortcut.lnk"
            _create_lnk_file(test_shortcut_path)
            link_created = test_shortcut_path.exists()
            if link_created:
                os.remove(test_shortcut_path)
            self.assertTrue(link_created)



#!/usr/bin/python
from pathlib import Path
import argparse
import sys

from .__init__ import __version__

parser = argparse.ArgumentParser(prog="txrm2tiff", description="Converter of txrm/xrm files to OME tif/tiff files", add_help=False)
convert_group = parser.add_argument_group("Converter arguments")
convert_group.add_argument('-i', '--input', dest='input_path', default=None, type=str, action='store', help='txrm or xrm file to convert, or directory containing txrm and/or xrm files')
convert_group.add_argument('-r', '--reference', dest='custom_reference', default=None, type=str, action='store', help='specify a custom file (tiff, txrm or xrm) for reference (ignores internal reference)')
convert_group.add_argument('-o', '--output-path', dest='output_path', default=None, action='store', help='specify output (can be directory or file) - specified directories that do not exist will be created')
convert_group.add_argument('-a', '--annotate', dest='annotate', action='store_true', help='save a second annotated copy, if annotations are available')
convert_group.add_argument('-d', '--datatype', dest='data_type', default=None, action='store', choices=['uint16', 'float32', 'float64'], help='specify output data type (default: decides data type from input)')
convert_group.add_argument('--ignore-ref', dest='ignore_reference', default=False, action='store_true', help='if specified ignore any internal reference')
convert_group.add_argument('--set-logging', dest='logging_level', default="info", type=str, action='store', help='pick logging level (options: debug, info, warning, error, critical)')

package_group = parser.add_argument_group("Package info")
package_group.add_argument('-v', '--version', action='version', version="%(prog)s {}".format(__version__))
package_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

subparsers = parser.add_subparsers(title="Additional options", help='Additional functionality options')
inspect_parser = subparsers.add_parser(name="inspect", add_help=False)
inspect_info_group = inspect_parser.add_argument_group("Inspection options")
inspect_info_group.add_argument('-i', '--input', dest='input_path', default="", type=str, action='store', help='txrm or xrm file to inspect')
inspect_info_group.add_argument('-e', '--extra', action='store_true', help="Shows additional file info")
inspect_info_group.add_argument('-l', '--list-streams', action='store_true', help="List all streams")
inspect_info_group.add_argument('-s', '--inspect-streams', dest='streams', nargs='+', help="Inspect stream")
inspect_info_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

setup_parser = subparsers.add_parser(name="setup", add_help=False)
setup_info_group = setup_parser.add_argument_group("Setup options")
setup_info_group.add_argument("-w", "--windows-shortcut", action='store_true', help="creates a Windows shortcut on the user's desktop that accepts drag and dropped files (allows batch processing)")
setup_info_group.add_argument('-h', '--help', action='help', help="show this help message and exit")


def main():
    if "inspect" in sys.argv:
        inspect_args = parser.parse_args(namespace=inspect_parser)
        if inspect_args.input_path:
            from .inspector import Inspector
            with Inspector(inspect_args.input_path) as inspector:
                inspector.inspect(extra=bool(inspect_args.extra))
                if inspect_args.list_streams:
                    inspector.list_streams()
                if inspect_args.streams:
                    inspector.inspect_streams(*inspect_args.streams)
                print(inspector.get_text())
                return
        else:
            inspect_parser.print_help()
            return

    if "setup" in sys.argv:
        setup_args = parser.parse_args(namespace=setup_parser)
        if setup_args.windows_shortcut:
            from .shortcut_creator import create_Windows_shortcut
            create_Windows_shortcut()
            return
        else:
            setup_parser.print_help()
            return

    args = parser.parse_args()
    if args.input_path:
        from .run import run
        run(**vars(args))

    else:
        parser.print_help()
        print("\n\n\ntxrm2tiff inspect:\n")
        inspect_parser.print_help()
        print("\n\n\ntxrm2tiff setup:\n")
        setup_parser.print_help()



import unittest
from unittest.mock import MagicMock, patch
from parameterized import parameterized

import os
import pathlib
from pathlib import Path
from random import randint
import string

from txrm2tiff.inspector import Inspector

visit_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/data/2019/cm98765-1")
raw_path = visit_path / "raw"
xm10_path = raw_path / "XMv10"
xm13_path = raw_path / "XMv13"

test_files = [
    (xm13_path / 'Xray_mosaic_v13.xrm', ),
    (xm13_path / 'Xray_mosaic_v13_interrupt.xrm', ),
    (xm13_path / 'Xray_single_v13.xrm', ),
    (xm13_path / 'tomo_v13_full.txrm', ),
    (xm13_path / 'tomo_v13_full_noref.txrm', ),
    (xm13_path / 'tomo_v13_interrupt.txrm', ),
    (xm13_path / 'VLM_mosaic_v13.xrm', ),
    (xm13_path / 'VLM_mosaic_v13_interrupt.xrm', ),
    (xm10_path / '12_Tomo_F4D_Area1_noref.txrm', ),
    (xm10_path / 'VLM_mosaic.xrm', ),
    (xm10_path / 'test_tomo2_e3C_full.txrm', ),
    (xm10_path / 'Xray_mosaic_F5A.xrm', ),
]

@unittest.skipUnless(visit_path.exists(), "dls paths cannot be accessed")
class TestInspectorFunctions(unittest.TestCase):

    @parameterized.expand(test_files)
    def test_basic_inspect(self, test_file):
        with Inspector(test_file) as inspector:
            inspector.inspect(False)
            text = inspector.get_text()
        self.assertNotIn("This is not a valid txrm/xrm file.", text)

    @parameterized.expand(test_files)
    def test_extra_inspect(self, test_file):
        with Inspector(test_file) as inspector:
            inspector.inspect(True)
            text = inspector.get_text()
        self.assertNotIn("This is not a valid txrm/xrm file.", text)

    @parameterized.expand(test_files)
    def test_list_streams(self, test_file):
        filename = Path(test_file).name
        with Inspector(test_file) as inspector:
            inspector.list_streams()
            # get all bits of text that aren't empty or the filename
            text_list = [i for i in inspector.get_text().split("\n") if i and i != filename]
            # There should be well over 10 lines
        self.assertTrue(len(text_list) > 10)


    @parameterized.expand(test_files)
    def test_read_streams(self, test_file):
        with Inspector(test_file) as inspector:
            stream_list = inspector._get_stream_list()
            indexes = [randint(0, len(stream_list) - 1) for _ in range(0, 10)]
            # Test a random assortment of keys
            for stream_key in [stream_list[i] for i in indexes]:
                inspector.inspect_streams(stream_key)
            text = inspector.get_text()
        self.assertNotIn("does not exist", text, msg="stream not read despite being in list")



import unittest
from unittest.mock import MagicMock, patch
from parameterized import parameterized

import numpy as np
from numpy.testing import assert_array_equal
from pathlib import Path

from txrm2tiff.txrm_to_image import TxrmToImage, save_colour
from txrm2tiff import annotator

test_data_path = Path("/dls/science/groups/das/ExampleData/B24_test_data/")
test_file = test_data_path / "annotation_test" / "Xray_mosaic_F3C.xrm"

visit_path = test_data_path / "data" / "2019" / "cm98765-1"
raw_path = visit_path / "raw"
xm10_path = raw_path / "XMv10"
xm13_path = raw_path / "XMv13"

test_files = [
    (test_file, ),
    (xm13_path / 'Xray_mosaic_v13.xrm', ),
    (xm13_path / 'Xray_mosaic_v13_interrupt.xrm', ),
    (xm13_path / 'Xray_single_v13.xrm', ),
    (xm13_path / 'tomo_v13_full.txrm', ),
    (xm13_path / 'tomo_v13_full_noref.txrm', ),
    (xm13_path / 'tomo_v13_interrupt.txrm', ),
    (xm13_path / 'VLM_mosaic_v13.xrm', ),
    (xm13_path / 'VLM_mosaic_v13_interrupt.xrm', ),
    (xm10_path / '12_Tomo_F4D_Area1_noref.txrm', ),
    (xm10_path / 'VLM_mosaic.xrm', ),
    (xm10_path / 'test_tomo2_e3C_full.txrm', ),
    (xm10_path / 'Xray_mosaic_F5A.xrm', ),]


class TestAnnotator(unittest.TestCase):

    @parameterized.expand(test_files)
    @unittest.skipUnless(visit_path.exists(), "dls paths cannot be accessed")
    def test_with_real_image(self, test_file):
        output_file = test_file.parent / (test_file.stem + "_Annotated.tif")

        converter = TxrmToImage()
        converter.convert(test_file, custom_reference=None, ignore_reference=False, annotate=True)
        ann_images = converter.get_annotated_images()
        self.assertFalse(ann_images is None, msg="Image wasn't created")
        save_colour(output_file, ann_images)
        self.assertTrue(output_file.exists(), msg=f"File {output_file} doesn't exist")
        output_file.unlink()

    @patch('txrm2tiff.annotator.txrm_wrapper.read_stream')
    def test_square(self, mocked_stream_reader):
        fill = 125
        im_size = 5
        im = np.zeros((1, im_size, im_size))

        x0, y0 = 1, 1
        x1, y1 = 4, 4

        ann = annotator.Annotator(im[0].shape[::-1])

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                mocked_stream_reader.side_effect = [[x0], [y0], [x1], [y1]]
                ole = MagicMock()
                stream_stem = ""
                ann._plot_rect(ole, stream_stem)
        ann._saved_annotations = True
        output = ann.apply_annotations(im)

        expected_output = np.zeros((im_size, im_size, 3), dtype=np.uint8)
        y0, y1 = im_size - 1 - y0, im_size - 1 - y1  # Invert y axis
        expected_output[y0, x0:x1 + 1, 1] = fill
        expected_output[y1:y0 + 1, x0, 1] = fill
        expected_output[y1, x0:x1 + 1, 1] = fill
        expected_output[y1:y0 + 1, x1, 1] = fill
        
        assert_array_equal(output[0], expected_output)

    @patch('txrm2tiff.annotator.txrm_wrapper.read_stream')
    def test_line(self, mocked_stream_reader):
        fill = 125
        x0, x1 = 1, 4
        y = 4
        im_size = 5
        im = np.zeros((1, im_size, im_size))

        ann = annotator.Annotator(im[0].shape[::-1])

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                mocked_stream_reader.side_effect = [[x0], [x1], [y], [y]]
                ole = MagicMock()
                stream_stem = ""
                ann._plot_line(ole, stream_stem)
        ann._saved_annotations = True
        output = ann.apply_annotations(im)

        expected_output = np.zeros((im_size, im_size, 3), dtype=np.uint8)
        y = im_size - 1 - y  # Invert y axis
        expected_output[y, x0:x1 + 1, 1] = fill

        assert_array_equal(output[0], expected_output)

    @patch('txrm2tiff.annotator.txrm_wrapper.read_stream')
    def test_ellipse(self, mocked_stream_reader):
        fill = 125
        xs = 0, 18
        ys = 2, 16
        x_mid = int(round(np.mean(xs)))
        y_mid = int(round(np.mean(ys)))
        im_size = 20
        im = np.zeros((1, im_size, im_size))

        ann = annotator.Annotator(im[0].shape[::-1])

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                mocked_stream_reader.side_effect = [[xs[0]], [ys[0]], [xs[1]], [ys[1]]]
                ole = MagicMock()
                stream_stem = ""
                ann._plot_ellipse(ole, stream_stem)
        ann._saved_annotations = True
        output = ann.apply_annotations(im)[0]

        self.assertTrue((output[:, :, 0] == 0).any())  # Red
        self.assertTrue((output[:, :, 2] == 0).any())  # Blue
        for (x, y) in ann._flip_y((x_mid, ys[0]), (xs[0], y_mid), (x_mid, ys[1]), (xs[1], y_mid)):
            self.assertEqual(output[y, x, 1], fill, msg="Failed for y, x of %i, %i" %(y, x))
    
    @patch('txrm2tiff.annotator.txrm_wrapper.read_stream')
    def test_polyline(self, mocked_stream_reader):
        fill = 125
        im_size = 18  # Must be an even number
        xs = tuple(range(im_size))
        ys = [0, im_size - 1] * (im_size // 2)
        total_points = len(xs)
        im = np.zeros((1, im_size, im_size))

        ann = annotator.Annotator(im[0].shape[::-1])

        with patch.object(ann, "_get_colour") as patched_colour:
            with patch.object(ann, "_get_thickness") as patched_thickness:
                patched_colour.return_value = (0, fill, 0, 255)
                patched_thickness.return_value = 1
                mocked_stream_reader.side_effect = [[total_points], xs, ys]
                ole = MagicMock()
                stream_stem = ""
                ann._plot_polyline(ole, stream_stem)
        ann._saved_annotations = True
        output = ann.apply_annotations(im)[0]

        self.assertTrue((output[:, :, 0] == 0).any())  # Red
        self.assertTrue((output[:, :, 2] == 0).any())  # Blue
        for x, y in ann._flip_y(*zip(xs, ys)):
            self.assertEqual(output[y, x, 1], fill)

    @patch('txrm2tiff.annotator.txrm_wrapper.read_stream')
    def test_scale_bar(self, mocked_stream_reader):
        im_size = 5
        im = np.zeros((1, im_size, im_size))

        ann = annotator.Annotator(im[0].shape[::-1])
        mocked_stream_reader.side_effect = [[17.]]
        ann._add_scale_bar(MagicMock())
        ann._saved_annotations = True
        output = ann.apply_annotations(im)[0]

        self.assertTrue((output[:, :, 0] == 0).any())    # Red
        self.assertTrue((output[:, :, 1] == 255).any())  # Green
        self.assertTrue((output[:, :, 2] == 0).any())    # Blue


    def test_flip_y(self):
        ann = annotator.Annotator((20, 20))
        res = ann._flip_y((0, 0), (19, 19))
        self.assertEqual(res, ((0, 19), (19, 0)))



import os
import logging
from pathlib import Path


dragndrop_bat_file = Path(__file__).resolve().parent / "scripts" / "dragndrop.bat"

def _get_Windows_home_path():
    home = os.getenv("USERPROFILE")
    if home is None:
        homedrive = os.getenv("HOMEDRIVE")
        homepath = os.getenv("HOMEPATH")
        if homedrive and homepath:
            home = os.path.join(homedrive, homepath)
        else:
            home = os.getenv("HOME")

    if home is not None and "Users" in home:
        try:
            home_path = Path(home).resolve(strict=True)
            return home_path
        except FileNotFoundError:
            pass
    logging.error("Cannot find valid home path. The following path was found: '%s'", home)
    raise FileNotFoundError(f"Cannot find valid home path. The following path was found: '{home}''")


def _create_lnk_file(shortcut_path):
    try:
        # win23com is from the package pywin32, only available in Windows
        import win32com.client
    except ImportError:
        msg = "win32com of pywin32 cannot be imported! Please run 'pip install pywin32' (with '--user' argument if on a shared python environment) then try again."
        print(msg)
        logging.error(msg)
        raise
    shell = win32com.client.Dispatch("WScript.Shell")
    shortcut = shell.CreateShortCut(str(shortcut_path))
    shortcut.Targetpath = str(dragndrop_bat_file)
    shortcut.save()
    msg = f"Desktop shortcut created! It can be found here: {shortcut_path}"
    print(msg)
    logging.info(msg)


def create_Windows_shortcut():
    if os.name != "nt":
        logging.error("This command is only valid on Windows installations.")
        return 
    # Place link on users desktop
    shortcut_path = _get_Windows_home_path() / "Desktop" / "txrm2tiff.lnk"
    msg = f"Creating shortcut on user desktop: {shortcut_path}"
    logging.info(msg)
    if shortcut_path.exists():
        msg = f"Existing txrm2tiff shortcut found:'{shortcut_path}'. Are you sure you want to replace it? (y/N)"
        user_input = str(input(msg))
        logging.debug(msg)
        logging.debug("User input: %s", user_input)
        if user_input.lower() == "y" or user_input.lower() == "yes":
            logging.info("The existing shortcut will be replaced.")
        elif user_input.lower() == "n" or user_input.lower() == "no":
            logging.info("The existing shortcut will not be modified.")
            return
        else:
            logging.info("Invalid input: %s. The existing shortcut will not be modified.", user_input)
            return
    _create_lnk_file(shortcut_path)
